# -*- coding: utf-8 -*-
from argparse import ArgumentParser
from psycopg2 import connect, ProgrammingError
import re


class PostgresRunner(object):
    parser = ArgumentParser('postgres')
    parser.add_argument(
        'uri',
        help='URI to the Postgres database to migrate'
    )

    uri_re = re.compile(
        r'^postgres://' \
        r'(?P<user>\S+?)(:(?P<password>\S+?))?@' \
        r'(?P<hostname>\S+?)?(:(?P<port>\d+))?/(?P<database>\S+)$'
    )

    def __init__(self, args):
        parsed = self.parser.parse_args(args)

        uri = self.uri_re.match(parsed.uri)
        if uri is None:
            # TODO: specify value format in error message
            raise ValueError('"%s" is not a valid postgres URI' % uri)
        params = uri.groupdict()

        self.conn = connect(
            database=params['database'],
            user=params['user'],
            password=params['password'],
            host=params['hostname'] or 'localhost',
            port=int(params['port'] or '5432'),
        )

        # stopped should be True if a migration raises an error
        self.stopped = False

    def _create_table(self):
        with self.conn.cursor() as cursor:
            cursor.execute('''
                CREATE TABLE IF NOT EXISTS trek_migrations (
                    name TEXT PRIMARY KEY
                )
            ''')

        self.conn.commit()


    def version(self):
        try:
            with self.conn.cursor() as cursor:
                cursor.execute('''
                    SELECT name FROM trek_migrations
                    ORDER BY name DESC
                    LIMIT 1
                ''')
                result = cursor.fetchone()

            self.conn.commit()

        except ProgrammingError as err:
            if 'relation "trek_migrations" does not exist' in str(err):
                self.conn.rollback()
                self._create_table()
            else:
                raise

            return self.version()
        finally:
            cursor.close()

        return result[0] if result is not None else result

    def _split_sql(self, sql):
        return [
            stmt for stmt in sql.split(';')
            if stmt.strip() != ''
        ]

    def up(self, name, migration):
        if self.stopped:
            yield 'info', 'Already stopped.'

        try:
            with self.conn.cursor() as cursor:
                map(cursor.execute, self._split_sql(migration.up))
                cursor.execute(
                    "INSERT INTO trek_migrations VALUES (%s)",
                    (name,)
                )

            self.conn.commit()

            yield 'info', 'Migrated %s up' % (name)
        except ProgrammingError as err:
            self.stopped = True
            yield 'error', 'Migrating %s up resulted in an error: %s' % (name, err)

    def down(self, name, migration):
        if self.stopped:
            yield 'info', 'Already stopped.'

        try:
            with self.conn.cursor() as cursor:
                map(cursor.execute, self._split_sql(migration.down))
                cursor.execute(
                    "DELETE FROM trek_migrations WHERE name = %s",
                    (name,)
                )

            self.conn.commit()

            yield 'info', 'Migrated %s down' % (name)
        except ProgrammingError as err:
            self.stopped = True
            yield 'error', 'Migrating %s down resulted in an error: %s' % (name, err)



# -*- coding: utf-8 -*-
from argparse import ArgumentParser, REMAINDER
from . import Migrator, ALL
import sys


parser = ArgumentParser('migrate')
parser.add_argument(
    '-c', '--count',
    type=int, default=-1,
    help='how many migrations to migrate',
)
parser.add_argument(
    '-r', '--runner',
    type=str, default='migrate.shell:ShellMigrator',
    help='which migrator to run (specify Python path)',
)
parser.add_argument(
    '-d', '--directory',
    type=str, default='migrations',
    help='which diretory to look for migrations in',
)
parser.add_argument(
    'direction',
    type=str, choices=['up', 'down'],
    help='which direction to migrate'
)
parser.add_argument(
    'rest', nargs=REMAINDER,
)

migrators = {
    'postgres': 'migrate.runners.postgres:PostgresRunner'
}

def display_messages(messages):
    for level, message in messages:
        if level == 'raw':
            print message
        else:
            print '[%s]: %s' % (level, message)

def run():
    args = parser.parse_args()

    if args.runner in migrators:
        args.runner = migrators[args.runner]

    if args.count < 0:
        if args.direction == 'up':
            args.count = ALL
        else:
            args.count = 1

    migrator = Migrator(
        args.count,
        args.runner,
        args.directory,
        args.direction,
        args.rest
    )

    try:
        display_messages(migrator.run())
    except (OSError, ValueError) as err:
        print '[error]: %s' % err
        sys.exit(1)

    sys.exit(0)



# -*- coding: utf-8 -*-
import itertools


class Migration(object):
    _up = 'MIGRATE UP'
    _down = 'MIGRATE DOWN'

    def _validate(self, content):
        """a migration should have *exactly* one up and down statement"""
        ups = content.count(self._up)
        if ups != 1:
            raise ValueError('Need exactly one up migration, have %d' % ups)

        downs = content.count(self._down)
        if downs != 1:
            raise ValueError('Need exactly one down migration, have %d' % downs)

    def _parse_up(self, content):
        start = list(itertools.dropwhile(
            lambda line: self._up not in line, content.split('\n')
        ))[1:]
        content = itertools.takewhile(
            lambda line: self._down not in line, start
        )
        return '\n'.join(content)

    def _parse_down(self, content):
        content = list(itertools.dropwhile(
            lambda line: self._down not in line, content.split('\n')
        ))[1:]
        return '\n'.join(content)

    def __init__(self, content):
        self._validate(content)
        self.up = self._parse_up(content)
        self.down = self._parse_down(content)



# -*- coding: utf-8 -*-
import pytest
from psycopg2 import connect
import os

@pytest.fixture(scope='session')
def postgres_info():
    return {
        'database': os.environ.get('PGDATABASE', 'trek_test'),
        'host': os.environ.get('PGHOST', 'localhost'),
        'port': int(os.environ.get('PGPORT', '5432')),
        'user': os.environ.get('PGUSER', 'trek_test'),
        'password': os.environ.get('PGPASSWORD', None),
    }

@pytest.yield_fixture(scope='session')
def connection(postgres_info):
    conn = connect(**postgres_info)
    yield conn
    conn.close()

@pytest.yield_fixture
def cursor(connection):
    cur = connection.cursor()
    yield cur
    cur.close()



# -*- coding: utf-8 -*-
from trek.migration import Migration
from trek.runners.postgres import PostgresRunner
from psycopg2 import ProgrammingError
import pytest

TABLE_QUERY = '''
SELECT tablename
FROM pg_tables
WHERE schemaname = 'public'
'''

TABLE = 'test1'
UP = 'CREATE TABLE %s (id INTEGER PRIMARY KEY)' % TABLE
DOWN = 'DROP TABLE %s' % TABLE

MIGRATION = Migration('-- MIGRATE UP\n%s\n-- MIGRATE DOWN\n%s' % (UP, DOWN))

def _create(connection):
    with connection.cursor() as cursor:
        cursor.execute(UP)
    connection.commit()

def _drop(connection):
    with connection.cursor() as cursor:
        try:
            cursor.execute(DOWN)
        except ProgrammingError as err:
            connection.rollback()
            if 'table "%s" does not exist' % TABLE in str(err):
                return
            else:
                raise

    connection.commit()

def _table_presence(connection, tablename, present):
    with connection.cursor() as cursor:
        cursor.execute(TABLE_QUERY)

        if present:
            assert (tablename,) in cursor.fetchall()
        else:
            assert (tablename,) not in cursor.fetchall()

@pytest.yield_fixture
def runner(postgres_info, connection, cursor):
    runner = PostgresRunner(['postgres://%s%s@%s%s/%s' % (
        postgres_info['user'],
        ':' + postgres_info['password'] if postgres_info['password'] else '',
        postgres_info['host'],
        ':%d' % postgres_info['port'] if 'port' in postgres_info else '',
        postgres_info['database'],
    )])
    assert not runner.stopped

    yield runner

    cursor.execute('DROP TABLE IF EXISTS trek_migrations')
    connection.commit()


class TestVersion(object):
    def test_creates_table(self, connection, runner):
        _table_presence(connection, 'trek_migrations', False)

        assert runner.version() is None

        _table_presence(connection, 'trek_migrations', True)

    def test_returns_version(self, connection, runner):
        runner._create_table()

        with connection.cursor() as cursor:
            cursor.execute("INSERT INTO trek_migrations VALUES ('1')")
        connection.commit()

        assert runner.version() == '1'


def test_up(request, connection, runner):
    runner._create_table()
    _table_presence(connection, TABLE, False)
    request.addfinalizer(lambda: _drop(connection))

    name = '1'
    messages = list(runner.up(name, MIGRATION))
    assert len(messages) == 1
    level, message = messages[0]

    assert level == 'info'
    assert message == 'Migrated 1 up'
    assert not runner.stopped

    with connection.cursor() as cursor:
        cursor.execute('SELECT name FROM trek_migrations')
        assert cursor.fetchall() == [(name,)]

    _table_presence(connection, TABLE, True)


def test_down(request, connection, runner):
    name = '1'
    runner._create_table()
    _create(connection)
    request.addfinalizer(lambda: _drop(connection))
    _table_presence(connection, TABLE, True)

    with connection.cursor() as cursor:
        cursor.execute("INSERT INTO trek_migrations VALUES (%s)", (name,))
    connection.commit()

    messages = list(runner.down(name, MIGRATION))
    assert len(messages) == 1
    level, message = messages[0]

    assert level == 'info'
    assert message == 'Migrated 1 down'
    assert not runner.stopped

    with connection.cursor() as cursor:
        cursor.execute('SELECT name FROM trek_migrations')
        assert cursor.fetchall() == []

    _table_presence(connection, TABLE, False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

def get(fname):
    try:
        with open(fname) as f:
            return f.read()
    except IOError:  # probably in tox
        return ''

setup(
    name='trek',
    version=get('VERSION'),
    description='generic migration provider',
    long_description=get('README.rst'),
    author='Brian Hicks',
    author_email='brian@brianthicks.com',
    url='https://github.com/brianhicks/trek',
    license=get('VERSION'),
    packages=find_packages(exclude=['tests']),

    entry_points={
        "console_scripts": [
            'trek = trek.cli:run'
        ]
    },
)



# -*- coding: utf-8 -*-
from .migration import Migration
from importlib import import_module
import os

# the way this is used in a slice means that if we want all the migrations we
# need to specify them as `None` (because `[1,2,3][:None] == [1,2,3]`). Since
# this is pretty confusing, it's specified in this name instead.
ALL = None


class Migrator(object):
    def __init__(self, count, runner_path, migrations_dir, direction, extra):
        self.count = count
        self.migrations_dir = migrations_dir
        self.direction = direction

        runner_cls = self.get_runner(runner_path)
        self.runner = runner_cls(extra)
        self.current = self.runner.version()

    def get_runner(self, path):
        package, name = path.split(':', 1)
        return getattr(import_module(package), name)

    def get_migration(self, name):
        with open(os.path.join(self.migrations_dir, name), 'r') as raw:
            return Migration(raw.read())

    def migrations_to_run(self):
        try:
            names = os.listdir(self.migrations_dir)
        except OSError:  # explicitly raising this. Deal with it!
            raise

        if not names:
            raise ValueError('No migrations to run in "%s"' % self.migrations_dir)

        if self.direction == 'up':
            return [
                name for name in names
                if self.current < name
            ][:self.count]
        elif self.direction == 'down':
            return [
                name for name in reversed(names)
                if self.current >= name
            ][:self.count]
        else:
            raise ValueError('Unknown migration direction "%s"' % self.direction)

    def migrate(self, names):
        if not names:
            yield 'info', 'No migrations necessary!'
            return

        for name in names:
            migration = self.get_migration(name)

            if self.direction == 'up':
                for message in self.runner.up(name, migration):
                    yield message
            elif self.direction == 'down':
                for message in self.runner.down(name, migration):
                    yield message
            else:
                raise ValueError(
                    'Unknown migration direction "%s"' % self.direction
                )

    def run(self):
        "put all the parts together"
        names = self.migrations_to_run()
        for pair in self.migrate(names):
            yield pair

        yield 'info', 'Ran %d migration(s)' % len(names)
